import netCDF4
import numpy as np
from obspy import Stream, Trace, UTCDateTime
from datetime import datetime, timezone
import tempfile
import shutil
import os

from geomagio import ChannelConverter
from geomagio.TimeseriesFactory import TimeseriesFactory
from geomagio.api.ws.Element import ELEMENT_INDEX


class NetCDFFactory(TimeseriesFactory):
    """Factory for reading/writing NetCDF format Geomagnetic Data using numeric epoch times."""

    # Define temperature element IDs
    TEMPERATURE_ELEMENTS_ID = {"UK1", "UK2", "UK3"}

    # Define the list of optional global attributes to handle
    OPTIONAL_GLOBAL_ATTRS = [
        "data_interval_type",
        "data_type",
        "station_name",
        "station",
        "network",
        "location",
        "declination_base",
        "sensor_orientation",
        "sensor_sampling_rate",
        "sample_period",
        "conditions_of_use",
    ]

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.time_format = "numeric"  # Fixed to numeric epoch times

    def _get_var_units_validmin_validmax(self, channel: str):
        """
        Determine units, valid_min, and valid_max based on channel name.

        Parameters
        ----------
        channel : str
            The channel name.

        Returns
        -------
        tuple
            A tuple containing (units, valid_min, valid_max).
        """
        channel = channel.upper()
        if channel == "D":
            units = "degree"
            valid_min = -360.0
            valid_max = 360.0
        elif channel == "I":
            units = "degree"
            valid_min = -90.0
            valid_max = 90.0
        # elif channel in self.TEMPERATURE_ELEMENTS_ID:
        #     units = "degree_Celsius"
        #     valid_min = -273.15
        #     valid_max = 79999.0
        elif channel in ["F", "S"]:
            units = "nT"
            valid_min = 0.0
            valid_max = 79999.0
        elif channel in ["X", "Y", "Z", "H", "E", "V", "G"]:
            units = "nT"
            valid_min = -79999.0
            valid_max = 79999.0
        else:
            units = getattr(channel, "units", "")
            valid_min = getattr(channel, "valid_min", np.nan)
            valid_max = getattr(channel, "valid_max", np.nan)
        return units, valid_min, valid_max

    def parse_string(self, data: bytes, **kwargs):
        """
        Parse a NetCDF byte string into an ObsPy Stream.

        Parameters
        ----------
        data : bytes
            Byte content of the NetCDF file.

        Returns
        -------
        Stream
            An ObsPy Stream object containing the parsed data.
        """
        try:
            nc_dataset = netCDF4.Dataset("inmemory", mode="r", memory=data)
        except Exception as e:
            print(f"Error loading NetCDF data: {e}")
            return Stream()

        # Map from known NetCDF attributes to the stats keys we actually use
        global_attrs_map = {
            "institution": "agency_name",
            "conditions_of_use": "conditions_of_use",
            "data_interval": "data_interval",
            "data_interval_type": "data_interval_type",
            "data_type": "data_type",
            "station_name": "station_name",
            "station": "station",
            "network": "network",
            "location": "location",
            "declination_base": "declination_base",
            "sensor_orientation": "sensor_orientation",
            "sensor_sampling_rate": "sensor_sampling_rate",
            "latitude": "geodetic_latitude",
            "longitude": "geodetic_longitude",
            "elevation": "elevation",
        }

        # Extract recognized global attributes
        global_attrs = {}
        for nc_attr_name, stats_attr_name in global_attrs_map.items():
            val = getattr(nc_dataset, nc_attr_name, None)
            if val is not None:
                global_attrs[stats_attr_name] = val

        # Convert "comment" to filter_comments array if present
        comment = getattr(nc_dataset, "comment", None)
        if comment:
            global_attrs["filter_comments"] = [comment]

        # Check for time variable
        if "time" not in nc_dataset.variables:
            print("No 'time' variable found in NetCDF data.")
            nc_dataset.close()
            return Stream()

        time_var = nc_dataset.variables["time"]
        times_epoch = time_var[:]

        # Collect the channels (all variables except 'time')
        channels = [v for v in nc_dataset.variables if v != "time"]
        channel_data = {}
        for ch in channels:
            channel_data[ch] = nc_dataset.variables[ch][:]

        # If no time points, return empty
        if len(times_epoch) == 0:
            print("No time points found in the NetCDF data.")
            return Stream()

        # Calculate sampling interval
        delta = (
            (times_epoch[-1] - times_epoch[0]) / (len(times_epoch) - 1)
            if len(times_epoch) > 1
            else 1.0
        )
        starttime = UTCDateTime(times_epoch[0])
        sampling_rate = 1.0 / delta if delta else 0

        stream = Stream()
        for channel_name, data_array in channel_data.items():
            # Length of data should matches times
            if len(data_array) != len(times_epoch):
                print(f"Channel '{channel_name}' has mismatched length. Skipping.")
                continue

            # Build stats from recognized attributes + basic fields
            stats = {}
            stats.update(global_attrs)  # put in all recognized global attrs

            # Basic per-trace stats
            stats["channel"] = channel_name
            stats["starttime"] = starttime
            stats["sampling_rate"] = sampling_rate
            # endtime, delta, npts, etc. will be handled automatically by ObsPy

            # Create Trace
            trace = Trace(data=np.array(data_array, dtype=float), header=stats)

            # Convert minutes to arcradians
            if channel_name in ["D", "I"]:
                if nc_dataset.variables[channel_name].units.startswith("deg"):
                    trace.data = ChannelConverter.get_minutes_from_degrees(trace.data)
                elif nc_dataset.variables[channel_name].units.startswith("rad"):
                    trace.data = ChannelConverter.get_minutes_from_radians(trace.data)
                elif nc_dataset.variables[channel_name].units.startswith("arc"):
                    pass  # nothing to do.

            stream.append(trace)

        nc_dataset.close()
        return stream

    def write_file_from_buffer(self, fh, timeseries: Stream, channels=None):
        """
        Write an ObsPy Stream to a NetCDF file, including optional attributes.

        Parameters
        ----------
        fh : str or file-like object
            File path or file handle to write the NetCDF data.
        timeseries : Stream
            ObsPy Stream object containing the data to write.
        channels : list, optional
            List of channel names to include. If None, all channels are included.
        """
        if not timeseries or len(timeseries) == 0:
            # Create an empty NetCDF structure with minimal CF compliance
            with netCDF4.Dataset(fh, "w", format="NETCDF4") as nc_dataset:
                nc_dataset.title = "Geomagnetic Time Series Data"
                nc_dataset.featureType = "timeSeries"  # per cf conventions
                nc_dataset.Conventions = "CF-1.6"
            return

        timeseries.merge()

        # Filter channels if specified
        if channels:
            channels = [ch.upper() for ch in channels]
            timeseries = Stream(
                [tr for tr in timeseries if tr.stats.channel.upper() in channels]
            )

        if len(timeseries) == 0:
            print("No matching channels found after filtering.")
            with netCDF4.Dataset(fh, "w", format="NETCDF4") as nc_dataset:
                nc_dataset.title = "Geomagnetic Time Series Data"
                nc_dataset.featureType = "timeSeries"  # per cf conventions
                nc_dataset.Conventions = "CF-1.6"
            return

        timeseries.sort(keys=["starttime"])
        tr = timeseries[0]
        stats = tr.stats

        # Extract necessary attributes
        lat = float(getattr(stats, "geodetic_latitude", 0.0))
        lon = float(getattr(stats, "geodetic_longitude", 0.0))
        alt = float(getattr(stats, "elevation", 0.0))
        reported_orientation = "".join(channels) if channels else ""

        npts = tr.stats.npts
        delta = tr.stats.delta
        starttime = tr.stats.starttime

        # Generate time values as epoch seconds
        times_epoch = np.array([starttime.timestamp + i * delta for i in range(npts)])

        # Create NetCDF dataset with CF conventions
        with netCDF4.Dataset(fh, "w", format="NETCDF4") as nc_dataset:
            # Define global attributes with CF compliance
            nc_dataset.Conventions = "CF-1.6"
            nc_dataset.title = "Geomagnetic Time Series Data"
            nc_dataset.featureType = "timeSeries"  # per cf conventions
            nc_dataset.institution = getattr(stats, "agency_name", "")
            # nc_dataset.source = getattr(stats, "source", "")
            history_entry = f"Created on {datetime.now(timezone.utc).isoformat()}"
            existing_history = getattr(nc_dataset, "history", "")
            if existing_history:
                nc_dataset.history = existing_history + "\n" + history_entry
            else:
                nc_dataset.history = history_entry
            nc_dataset.references = getattr(stats, "references", "")
            nc_dataset.comment = getattr(stats, "filter_comments", "")

            # Add optional global attributes if they exist in stats and are not empty
            for attr in self.OPTIONAL_GLOBAL_ATTRS:
                val = getattr(stats, attr, None)
                if val not in [None, ""]:
                    setattr(nc_dataset, attr, val)

            if reported_orientation:
                setattr(nc_dataset, "reported_orientation", reported_orientation)

            # Define the time dimension
            nc_dataset.createDimension("time", npts)

            # Create the time variable
            time_var = nc_dataset.createVariable("time", "f8", ("time",))
            time_var.units = "seconds since 1970-01-01T00:00:00Z"  # CF-compliant units
            time_var.calendar = getattr(stats, "calendar", "standard")
            time_var.standard_name = "time"
            time_var.long_name = "Time"
            time_var.axis = "T"
            time_var[:] = times_epoch

            # Assign valid range if available
            if "valid_min" in stats and "valid_max" in stats:
                time_var.valid_min = stats["valid_min"]
                time_var.valid_max = stats["valid_max"]

            # Create channel variables with CF attributes
            for trace in timeseries:
                ch_name = trace.stats.channel.upper()
                # Convert arcminutes to radians
                if "D" == ch_name:
                    trace.data = ChannelConverter.get_degrees_from_minutes(trace.data)

                if len(ch_name) > 63:
                    raise ValueError(
                        f"Channel name '{ch_name}' exceeds NetCDF variable name length limit."
                    )

                # Ensure channel name is a valid NetCDF variable name
                ch_name_nc = ch_name.replace("-", "_").replace(" ", "_")

                # Determine units, valid_min, and valid_max based on channel
                units, valid_min, valid_max = self._get_var_units_validmin_validmax(
                    ch_name
                )

                var = nc_dataset.createVariable(
                    ch_name_nc,
                    "f8",
                    ("time",),
                    fill_value=getattr(trace.stats, "_FillValue", 99_999),
                )

                # Set CF attributes
                var.units = units
                if ELEMENT_INDEX.get(ch_name):
                    # remove arcminutes, other things
                    long_name = ELEMENT_INDEX.get(ch_name).name.replace(
                        "(arcminutes)", ""
                    )
                    long_name = long_name.strip()
                    var.long_name = getattr(trace.stats, "long_name", long_name)
                var.standard_name = getattr(
                    trace.stats, "standard_name", f"geomagnetic_field_{ch_name.lower()}"
                )

                # Set data validity attributes
                var.valid_min = valid_min
                var.valid_max = valid_max

                # Assign data
                var[:] = trace.data

            # Define geospatial coordinates
            nc_dataset.latitude = lat
            # nc_dataset.lat_units = "degrees_north"
            nc_dataset.longitude = lon
            # nc_dataset.geospatial_lon_units = "degrees_east"
            nc_dataset.elevation = alt
            # nc_dataset.elevation_units = "meters"

    def write_file(self, fh, timeseries: Stream, channels=None):
        """
        Write an ObsPy Stream to a NetCDF file.

        Parameters
        ----------
        fh : file-like object
            File handle to write the NetCDF data.
        timeseries : Stream
            ObsPy Stream object containing the data to write.
        channels : list, optional
            List of channel names to include. If None, all channels are included.
        """
        # Create a temporary file
        with tempfile.NamedTemporaryFile(delete=False) as tmp:
            tmp_filepath = tmp.name

        try:
            # Write to the temporary file
            self.write_file_from_buffer(tmp_filepath, timeseries, channels)

            # Read the temporary file content into destination file
            with open(tmp_filepath, "rb") as tmp:
                shutil.copyfileobj(tmp, fh)
        finally:
            # Clean up the temporary file
            os.remove(tmp_filepath)
